/*
**++
**  FACILITY:
**
**      thermo.c
**
**  ABSTRACT:
**
**      This program reads and logs temperatures. It communicates with
**      a digital meter via serial link to read voltage. The voltage
**      temperatur coefficient is assumed to be 10 mV/degree F.
**
**      Usage:
**
**          thermo [-option] ComPortNumber CollectInterval [OutputPath]
**
**          Where option is one of d for debugging, l for event logging,
**          or la for log all events.
**
**          Debugging bypasses the actual read, and uses fixed data.
**          It also turns on logging of all events.
**
**          Logging logs events of interest to a log file named THERMO.LOG.
**
**          Log all logs even boring events to THERMO.LOG.
**
**          ComPortNumber is the number of the COM port to use for
**          communications with the meter, 1 for COM1, 2 for COM2.
**
**          CollectInterval is the interval, in seconds, between collection
**          events.
**
**          The output is stored in a file named YYMMDD.DAT where YY is the
**          year of the century, MM is the month number (01 for January,
**          02 for February, etc.), and DD is the day of the month.
**
**          OutputPath contains an optional path in which to store the data.
**          This can include a disc letter followed by a colon if desired.
**
**          Once per hour, the data and log files are flushed by closing
**          and reopening.
**
**          Once per day the program calls the program FIXTIME which reads
**          the hardware time of day clock and sets DOS time of day.
**
**          At any time one may press a key to request special actions.
**          The keys and actions one may request include (case irrelevant)
**
**              F   flush data and log files by closing and reopening
**              V   display version information
**              Esc exit the program
**
**          where Esc is the escape key.
**
**  AUTHORS:
**
**      J.M.McCarty
**
**  CREATION DATE:    10/??/1997
**
**  MODIFICATION HISTORY:
**
**      Date        Version History
**      ----        ------- -------
**      12 Jun 1999 10.1    Added standard file and function prologues
**                          Started history
**                          Made -l only log interesting actions
**                          Added -la to log all actions
**                          Made program display all UNKNOWN actions to stderr
**                              (used to be Action == UNKNOWN, now it is
**                                          Action & UNKNOWN)
**                          Added ReportFailure
**                          Added ReportExit
**                          Now logs failures to log file
**                          Now logs exits to log file and to stderr
**                          Added explicit close of log file on exit
**                          Added GetTemperature to insulate GetDCVoltage
**                              from its environment
**                          Added ReportVersion
**                          Added GetTime
**                          Changed MustOpen(.,.) to call ReportError
**
**      20 Jun 1999 10.2    Added GetTimeOfDay(.,.). On certain occasions the
**                          clock interrupt would occur just between getting
**                          epoch time and getting the clock or wall time.
**                          This would result in starting a new day without
**                          doing a collection when it was needed.
**                          GetTimeOfDay(.,.) ensures that the epoch time
**                          and wall time are in synchronization.
**
**       8 Apr 2000 10.3    Changed DEFER_INTERVAL from 8 seconds to 20
**                          seconds, because the machine used for logging
**                          had a screen failure. This gives more time for
**                          changing the floppy used to store the data.
**--
**/

/*
**
**  INCLUDE FILES
**
**/

#include <ctype.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>     /* strtod(.,.) */
#include <time.h>

#include <dos.h>        /* gettime(.) */
#include <io.h>         /* isatty(.) */

#include "meter.h"


/*
**
**  MACRO DEFINITIONS
**
**/

#define VERSION_MAJOR   10
#define VERSION_MINOR   3

/* Action Bits */

#define NOTHING         0x00
#define COLLECT         0x01
#define FLUSH           0x02
#define NEW_DAY         0x04
#define EXIT            0x08
#define VERSION         0x10
#define UNKNOWN         0x20

#define SECOND  1
#define MINUTE  (60*SECOND)
#define HOUR    (60*MINUTE)

#define DEFER_INTERVAL          (20*SECOND)
#define FLUSH_INTERVAL          (1*HOUR)
#define DEBUG_FLUSH_INTERVAL    (5*SECOND)

#define CONTROL(x)  ((x)-0x40)
#define ESC_KEY     CONTROL('[')

/*
**
**  TYPEDEFS
**
**/

typedef enum    { False, True } Boolean;

Boolean Debugging = False;

/*
**
**  EXTERNAL DEFINITIONS
**
**/


/*
**
**  GLOBAL AND STATIC DEFINITIONS
**
**/

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      ReportError     --  report an error and die with error code
**
**  FORMAL PARAMETERS:
**
**      Format  input   format string a la printf for error display
**      ...     input   arguments to format and display to stderr
**
**  IMPLICIT INPUTS:
**
**      none
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      none
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


void    ReportError(char *Format,...) {
    va_list Args;

    va_start(Args,Format);
    vfprintf(stderr,Format,Args);
    va_end(Args);
    fprintf(stderr,"\n");
    exit(EXIT_FAILURE);
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      GetDCVoltage    --  query the meter for DC voltage
**
**  FORMAL PARAMETERS:
**
**      Voltage output  voltage read if not debugging, else 1.0
**
**  IMPLICIT INPUTS:
**
**      none
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      0 for failure, 1 for success
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


Boolean GetDCVoltage(double *Voltage) {
    int     Result;
    char    Reading[METER_READING_SIZE], *Value, *EndPtr;
    double  Multiplier;

    Result = False;
    if (GetMeterReading(Reading)
    &&  Reading[0] == 'D' && Reading[1] == 'C') {
        Value = Reading+2;
        while (*Value == ' ')
            Value++;
       *Voltage = strtod(Value,&EndPtr);
        while (*EndPtr == ' ')
            EndPtr++;
        switch (*EndPtr) {
            case 'V':   Multiplier =       1.0;             break;
            case 'm':   Multiplier =     0.001; EndPtr++;   break;
            default:    Multiplier =      -1.0;             break;
        }
        if (Multiplier > 0.0) {
           *Voltage *= Multiplier;
            if (strcmp(EndPtr,"V") == 0)
                Result = True;
        }
    }
    return Result;
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      GetTemp --  get temperature
**
**  FORMAL PARAMETERS:
**
**      TempF   output  temperature Fahrenheit
**      TempC   output  temperature Celsius
**
**  IMPLICIT INPUTS:
**
**      Debugging
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      True    successful
**      False   unsuccessful
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/

Boolean GetTemperature(double *TempF,double *TempC) {
    Boolean Result;
    double  Voltage;

    if (Debugging) {
        Result = True;
        Voltage = 1.0;
    } else
        Result = GetDCVoltage(&Voltage);
    if (Result) {
       *TempF = 100.0*Voltage;
       *TempC = (*TempF-32.0)*5.0/9.0;
    }
    return Result;
}


/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      FixTime     --  invoke the FIXTIME program
**
**  FORMAL PARAMETERS:
**
**      none
**
**  IMPLICIT INPUTS:
**
**      none
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      none
**
**  SIDE EFFECTS:
**
**      Changes the system time to conform to the hardware real time clock.
**
**--
**/


void    FixTime(void){
    system("fixtime > nul");
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      NextTime    --  given an interval, determine the next epoch at
**                      which an event comes due
**
**  FORMAL PARAMETERS:
**
**      Interval    input   interval at which events come due
**      Offset      input   flag indicating whether to skip the current
**                          epoch if it happens to coincide with an event
**
**  IMPLICIT INPUTS:
**
**      system time
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      The next epoch at which an event comes due.
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


time_t  NextTime(int Interval,Boolean Offset) {
    time_t  Result;
    ldiv_t  Ldiv;

    Result = time(NULL);
    Ldiv = ldiv(Result,Interval);
    if (Offset || Ldiv.rem)
        Ldiv.quot++;
    Result = Ldiv.quot*Interval;
    return Result;
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      DisplayAction   --  display the list of actions due
**
**  FORMAL PARAMETERS:
**
**      Output  inout   file on which to display the actions
**      Time    input   current time in string format
**      Action  input   list of actions due
**
**  IMPLICIT INPUTS:
**
**      none
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      none
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


void    DisplayAction(FILE *Output,const char *Time,int Action) {
    Boolean OnePrinted;
    typedef struct {
        int     ActionBit;
        char   *ActionName;
    } ActionName_t;
    ActionName_t    ActionTable[] = {
        { COLLECT,  "COLLECT" },
        { FLUSH,    "FLUSH"   },
        { NEW_DAY,  "NEW_DAY" },
        { EXIT,     "EXIT"    },
        { VERSION,  "VERSION" },
        { UNKNOWN,  "UNKNOWN" },
        { NOTHING,  "NOTHING" }
    };
    ActionName_t   *Entry;

    if (Time) {
        fputs(Time,Output);
        fputs(" ",Output);
    }
    fputs("Action = ",Output);
    OnePrinted = False;
    for (Entry = ActionTable; Entry->ActionBit; Entry++) {
        if (Action & Entry->ActionBit) {
            if (OnePrinted)
                fputc('+',Output);
            fputs(Entry->ActionName,Output);
            OnePrinted = True;
        }
    }
    if (!OnePrinted)
        fputs("NOTHING",Output);
    fputc('\n',Output);
}
            
/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      GetTimeOfDay    --   get time and wall time
**
**  FORMAL PARAMETERS:
**
**      Time        output  epoch time
**      WallTime    output  clock time
**
**  IMPLICIT INPUTS:
**
**      none
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      none
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/

void    GetTimeOfDay(time_t *Time,struct time *WallTime) {
    time_t  ThisTime, PrevTime;

    /* On occasion the time will change between the call to   */
    /* time(.) and gettime(.). This loop ensures that the two */
    /* are in synchronization.                                */
    ThisTime = time(NULL);
    do {
        PrevTime = ThisTime;
        gettime(WallTime);
        ThisTime = time(NULL);
    } while (ThisTime != PrevTime);
   *Time = ThisTime;
}


/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      WaitForEvent    --  wait for next event to come due
**
**  FORMAL PARAMETERS:
**
**      CollectInterval input   interval between collection events
**      FlushInterval   input   interval between flush events
**
**  IMPLICIT INPUTS:
**
**      system time
**      Debugging       debugging flag
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      List of events due.
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


int     WaitForEvent(int CollectInterval,int FlushInterval) {
    int             Result, PrevResult;
    time_t          Time;
    static time_t   NextFlushTime, NextCollectTime;
    struct time     WallTime;
    static Boolean  FirstCall = True, NewDayDone = False;
    Boolean         Deferring, Forced;

    if (FirstCall) {
        NextCollectTime = NextTime(CollectInterval,False);
        NextFlushTime = NextTime(FlushInterval,False);
        FirstCall = False;
    }

    Forced = False;
    Deferring = False;
    PrevResult = Result = NOTHING;
    while (Result == NOTHING
    ||    (Deferring && !(Result & COLLECT))
    ) {
        GetTimeOfDay(&Time,&WallTime);
        if (Time >= NextCollectTime)
            Result |= COLLECT;
        if (Time >= NextFlushTime)
            Result |= FLUSH;
        if (WallTime.ti_hour == 0 && !NewDayDone)
            Result |= NEW_DAY;
        if (WallTime.ti_hour != 0)
            NewDayDone = False;
        Forced = kbhit();
        if (Forced) {
            switch (toupper(getch())) {
                case ESC_KEY:   Result |= EXIT;     break;
                case 'F':       Result |= FLUSH;    break;
                case 'V':       Result |= VERSION;  break;
                default:        Result |= UNKNOWN;  break;
            }
        }
        if ((Result & ~COLLECT) && (NextCollectTime-Time < DEFER_INTERVAL)
        && (!Deferring || Result != PrevResult)) {
            if (Forced || Debugging)
                fprintf(stderr,"Deferring action to next collection period.\n");
            if (Debugging) {
                DisplayAction(stderr,NULL,Result);
                fprintf(stderr,"Time = %ld, next collection period = %ld\n",
                    Time,NextCollectTime);
            }
            Deferring = True;
            PrevResult = Result;
        }
    }
    if (Result & COLLECT)
        NextCollectTime = NextTime(CollectInterval,True);
    if (Result & FLUSH)
        NextFlushTime = NextTime(FlushInterval,True);
    if (Result & NEW_DAY) {
        NewDayDone = True;
        Result &= ~FLUSH;
    }
    return Result;
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      MustOpen    --  open a file or die trying
**
**  FORMAL PARAMETERS:
**
**      FileName    input   complete path name of the file
**      Mode        input   mode for the open (see fopen)
**
**  IMPLICIT INPUTS:
**
**      Debugging       debugging flag
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      Pointer to open file.
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


FILE   *MustOpen(const char *FileName,const char *Mode) {
    FILE   *Result;

    if (Debugging)
        fprintf(stderr,"Opening \"%s\" with mode \"%s\"\n",FileName,Mode);
    Result = fopen(FileName,Mode);
    if (Result == NULL)
        ReportError("Unable to open \"%s\" with mode \"%s\"",FileName,Mode);
    return Result;
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      MakeFullPath    --  given an input path and file name create a full
**                          path name to the file.
**
**  FORMAL PARAMETERS:
**
**      FullPath    output  full path name of the file
**      Path        input   path name to be prepended to the file
**      FileName    input   bare file name
**
**  IMPLICIT INPUTS:
**
**      none
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      Pointer to the full path name.
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


char   *MakeFullPath(char *FullPath,const char *Path,const char *FileName) {
    int     Len;
    
    Len = strlen(Path);
    strcpy(FullPath,Path);
    if (*Path && Path[Len-1] != '\\' && Path[Len-1] != '/' && Path[Len-1] != ':')
        strcat(FullPath,"\\");
    strcat(FullPath,FileName);
    return FullPath;
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      OpenOutput  --  build the file name of the data file and open it.
**
**  FORMAL PARAMETERS:
**
**      Path    input   path to prepend to the (created) file name
**
**  IMPLICIT INPUTS:
**
**      system time
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      Pointer to the opened file.
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


FILE   *OpenOutput(const char *Path) {
    char    FileName[FILENAME_MAX];
    char    FullPath[FILENAME_MAX];
    time_t  Time;

    Time = time(NULL);
    strftime(FileName,sizeof(FileName)-1,"%y%m%d.dat",localtime(&Time));
    MakeFullPath(FullPath,Path,FileName);
    return MustOpen(FullPath,"at");
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      OpenLog --  build the file name for the log file and open it.
**
**  FORMAL PARAMETERS:
**
**      Path    input   path to prepend to the (fixed) file name
**
**  IMPLICIT INPUTS:
**
**      none
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      Pointer to the opened file.
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


FILE   *OpenLog(const char *Path) {
    char    FullPath[FILENAME_MAX];

    MakeFullPath(FullPath,Path,"thermo.log");
    return MustOpen(FullPath,"at");
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      Flush   --  flush the data file by closing and reopening it.
**
**  FORMAL PARAMETERS:
**
**      Output  inout   file to flush
**      Path    input   optional path to prepend
**
**  IMPLICIT INPUTS:
**
**      none
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      none
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


void    Flush(FILE **Output,FILE *(*Open)(const char *),const char *Path) {
    fclose(*Output);
   *Output = Open(Path);
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      Usage   --  report usage error and die.
**
**  FORMAL PARAMETERS:
**
**      none
**
**  IMPLICIT INPUTS:
**
**      none
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      none
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


void    Usage(void) {
    ReportError(
        "Usage: thermo [-option] ComPortNumber CollectInterval [OutputPath]\n"
        "Where option is one of d for debugging or l for logging, la for log all."
    );
}
/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      GetArgs --  parse the input command line and get the program
**                  control arguments.
**
**  FORMAL PARAMETERS:
**
**      Debugging       output  -d was encountered
**      Logging         output  -d, -l, or -la was encountered
**      LogAll          output  -d or -la was encountered
**      PortNumber      output  COM port number to use
**      CollectInterval output  collection interval in seconds
**      FlushInterval   output  flush interval in seconds
**      Path            output  optional path to prepend to data and log file
**                              names
**      Argc            input   argument count (a la ANSI main())
**      Argv            input   argument values (a la ANSI main())
**
**  IMPLICIT INPUTS:
**
**      none
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      none
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


void    GetArgs(Boolean *Debugging,Boolean *Logging,Boolean *LogAll,
        int *PortNumber, int *CollectInterval, int *FlushInterval,char **Path,
        int Argc,char *Argv[]) {
    int     ArgNum;

    if (Argc < 3 && Argc > 5)
        Usage();
    ArgNum = 1;
   *Debugging = False;
   *Logging = False;
   *LogAll = False;
    if (!strcmp(Argv[ArgNum],"-d")) {
       *Debugging = True;
       *Logging = True;
       *LogAll = True;
        ArgNum++;
    } else if (!strcmp(Argv[ArgNum],"-la")) {
       *Logging = True;
       *LogAll = True;
        ArgNum++;
    } else if (!strcmp(Argv[ArgNum],"-l")) {
       *Logging = True;
        ArgNum++;
    }
   *PortNumber = atoi(Argv[ArgNum++]);
   *CollectInterval = atoi(Argv[ArgNum++]);
   *FlushInterval = (*Debugging) ? (DEBUG_FLUSH_INTERVAL) : (FLUSH_INTERVAL);
   *Path = (ArgNum < Argc) ? Argv[ArgNum++] : ".";
    if (ArgNum != Argc)
        Usage();
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      ReportFailure   --  report failure to receive response from meter
**
**  FORMAL PARAMETERS:
**
**      File    inout   file to report the failure on
**      Time    input   time of day as string
**
**  IMPLICIT INPUTS:
**
**      none
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      none
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


void    ReportFailure(FILE *File,const char *Time) {
    fprintf(File,"%s Failed to receive.\n",Time);
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      ReportVersion   --  report version and run parameter information
**
**  FORMAL PARAMETERS:
**
**      Time            input   current time as a string
**      PortNumber      input   COM port number
**      CollectInterval input   collection interval
**
**  IMPLICIT INPUTS:
**
**      none
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      none
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


void    ReportVersion(FILE *File,const char *Time,int PortNumber,
            int CollectInterval) {
    if (Time != NULL)
        fprintf(File,"%s ",Time);
    fprintf(File,
        "THERMO %d.%d -- Getting Temperature Readings from COM%d at %d second intervals.\n",
        VERSION_MAJOR,VERSION_MINOR,PortNumber,CollectInterval);
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      ReportExit  --  report exit from program
**
**  FORMAL PARAMETERS:
**
**      File    inout   file to report the exit on
**      Time    input   time of day as string
**
**  IMPLICIT INPUTS:
**
**      none
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      none
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


void    ReportExit(FILE *File,const char *Time) {
    fprintf(File,"%s Exiting program.\n",Time);
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      GetTime --  get time formatted
**
**  FORMAL PARAMETERS:
**
**      TimeBuf output  time buffer (assumed to have at least 25 characters)
**
**  IMPLICIT INPUTS:
**
**      system time
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      Pointer to formatted time of day.
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


char   *GetTime(char *TimeBuf) {
    time_t  Time;

    Time = time(NULL);
    strcpy(TimeBuf,ctime(&Time));
    TimeBuf[24] = '\0';
    return TimeBuf;
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      main    --  main entry point
**
**  FORMAL PARAMETERS:
**
**      Argc    input   argument count
**      Argv    input   argument values
**
**  IMPLICIT INPUTS:
**
**      system time
**
**  IMPLICIT OUTPUTS:
**
**      none
**
**  FUNCTION VALUE:
**
**      EXIT_SUCCESS    normal termination
**      EXIT_FAILURE    abnormal termination
**
**  SIDE EFFECTS:
**
**      none
**
**--
**/


int     main(int Argc,char *Argv[]) {
    char    TimeBuf[26];
    int     PortNumber;
    double  Voltage, TemperatureC, TemperatureF;
    int     CollectInterval, FlushInterval;
    char   *Path;
    FILE   *Output, *Log;
    int     OutputIsTTY;
    int     Action;
    Boolean Logging, LogAll;

    GetArgs(&Debugging,&Logging,&LogAll,&PortNumber,&CollectInterval,
        &FlushInterval,&Path,Argc,Argv);

    ReportVersion(stderr,NULL,PortNumber,CollectInterval);
    fprintf(stderr,"\n");
    fprintf(stderr,"Press ESC to exit, F to flush output, or V to get version.\n");
    fprintf(stderr,"\n");

    Output = OpenOutput(Path);
    OutputIsTTY = isatty(fileno(Output));

    /* Call GetTime(.) again in case user stopped output with ^S */
    if (Logging) {
        Log = OpenLog(Path);
        ReportVersion(Log,GetTime(TimeBuf),PortNumber,CollectInterval);
    }

    OpenMeter(PortNumber);

    for (;;) {
        Action = WaitForEvent(CollectInterval,FlushInterval);
        GetTime(TimeBuf);
        if (Logging && (LogAll || Action != COLLECT))
            DisplayAction(Log,TimeBuf,Action);
        if (Debugging || Action == NOTHING || (Action & UNKNOWN))
            DisplayAction(stderr,TimeBuf,Action);
        if (Action & NEW_DAY) {
            fprintf(stderr,"%s Starting new day.\n",TimeBuf);
            Flush(&Output,OpenOutput,Path);
            if (Logging)
                Flush(&Log,OpenLog,Path);
            FixTime();
        }
        if (Action & COLLECT) {
            if (!GetTemperature(&TemperatureF,&TemperatureC)) {
                ReportFailure(Output,TimeBuf);
                if (!OutputIsTTY)
                    ReportFailure(stderr,TimeBuf);
                if (Logging)
                    ReportFailure(Log,TimeBuf);
            } else
                fprintf(Output,"%s %.1f F\n",TimeBuf,TemperatureF);
        }
        if (Action & FLUSH) {
            fprintf(stderr,"%s Forcing output to disc.\n",TimeBuf);
            Flush(&Output,OpenOutput,Path);
            if (Logging)
                Flush(&Log,OpenLog,Path);
        }
        if (Action & VERSION)
            ReportVersion(stderr,TimeBuf,PortNumber,CollectInterval);
        if (Action & EXIT) {
            ReportExit(stderr,TimeBuf);
            if (Logging)
                ReportExit(Log,TimeBuf);
            break;
        }
    }
    fclose(Output);
    if (Logging)
        fclose(Log);
    CloseMeter();
    return EXIT_SUCCESS;
}
